!+
! Module: hkauxplot_mod
!
! Authors:  Brendan Perry - University of Leicester
!
! Description: plots selected HK parameters to selected output devices
!
! Comments:
!
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!-

module hkauxplot_mod
  
  use dal
  use oal
  use datafilter
  use param
  use errorhandling
  use arrayeval
  
  implicit none
  
  private
  
  ! task parameter names 
  
  character(len=*), parameter :: useodfpar = "useodf" , &
                                 infilepar = "set" , &
                                 usexblkpar = "useblock" , &
                                 xblkpar = "block" , &
                                 xcolpar = "xcolumnname" , &
                                 colnamespar = "ycolumnnames", &
                                 usegtipar = "usegtiset" , &
                                 gtisetpar = "gtiset" , &
                                 useccdspar = "useccds" , &
                                 selccdpar = "selectccds" , &
                                 ccdpar = "ccds" , &
                                 devicepar = "plotdevice" , &
                                 selectoutpar = "selectoutfile" , &
                                 outfilepar = "outfile" , &
                                 offsetpar = "offsettime" , &
                                 selrowspar = "selectrows" , &
                                 startRpar = "startrow" , &
                                 stopRpar = "stoprow" , &
                                 seltimepar = "selecttimes" , &
                                 startTpar = "starttime" , &
                                 stopTpar = "stoptime" , &
                                 selframepar = "selectframes" , &
                                 startFpar = "startframe" , &
                                 stopFpar = "stopframe" , &
                                 pointspar = "points" , &
                                 convertpar = "convertframe"

  character(len=8), parameter :: hkdatatype = "PERHK.EL" , &
                                 auxdatatype = "AUX.EL" , &
                                 gtidatatype = "STDGTI" , &
                                 startcolpar = "START" , & 
                                 stopcolpar = "STOP" , &
                                 dtpar = "DATATYPE" , &
                                 timetype = "TIME" , &
                                 frametype = "FRAME" , &
                                 quadranttype = "QUADRANT" , &
                                 extname = "PERIODIC" , &
                                 ccdcol = "CCDID"

  character(len=80) :: telescop, instrume, obs_id, exp_id, &
                       date_obs , date_end , object
  character(256), dimension(7) :: labell
  real(single) :: tstartin , tstopin , mjdref , duration , &
                  start , stop
  integer :: nccds , ccds , row1 , rown
  logical :: useodf , usexblk , hk , aux , other , frames , gtiyes , &
             useccds, timeyes , selectoutfile ,  convert  , selectccds , &
             selectrows, selecttimes, selectframes, getstartstop, &
             getranges

  public hkauxplot
  
contains
  
  !---------------------------------------------------------------------
  subroutine hkauxplot
    !---------------------------------------------------------------------
    
    character(len=1024) :: infile , gtifile , plotfile
    character(len=80) :: device , xblockname , xcolumnname , xcolunit
    integer :: i , j , nycols , ninrows, ngoodcols , points , ccdno , &
               maxgtirows
    logical :: offset
    !    real(single) :: prevtstartin , prevtstopin
    real(single), allocatable, dimension(:) :: temp_gti , temp_x , temp
    integer, allocatable, dimension(:) :: minlox,minloy,maxlox,maxloy
    !    
    ! table and set declarations
    !
    type(DataSetT) :: in_set
    type(TableT)   :: in_tab
    !
    ! pointer declarations
    !
    character(len=80), dimension(:), pointer ::   ycolumnsptr,goodcolsptr,&
                                                  ycolunitsptr
    integer(kind=Int16), dimension(:), pointer :: ccd_ptr
    integer(kind=Int32), dimension(:), pointer :: n_ccd_row , n_gti_row
    real(kind=double), dimension(:), pointer :: xmin_ptr, xmax_ptr
    real(kind=double), dimension(:,:), pointer :: x_out_ptr, ycolrowptr, &
                                                  gti_array, gti_out
    real(kind=double), dimension(:,:,:), pointer :: y_out_ptr, gti_ccd_ptr
    !
    ! variable initialization
    !
    hk = .FALSE.        ! is file a housekeeping file?
    aux = .FALSE.       ! is file an auxiliary file?
    frames = .FALSE.    ! does file contain frame data?
    other = .FALSE.     ! is file not hk or aux?
    convert = .FALSE.   ! convert FRAME to time real(single)?
    gtiyes = .FALSE.    ! apply gti's?
    offset = .FALSE.    ! offset times (time = time - time(0))?
    timeyes = .FALSE.   ! x-axis is TIME-type
    getranges = .FALSE. ! ranges of rows/times/frames not selected?
    nccds = 1
    ccds = 1

    ! Get all the parameters first in subroutine getparamvalues
    
    call message(AppMsg,Sparse,"Reading parameter file...")

    call getparams(infile, xblockname, xcolumnname, ycolumnsptr, &
                   gtifile, nycols, ccd_ptr, device, plotfile, &
                   points, offset)
    
    ! Check that the in file exists, and has at least SOME of the
    ! columns requested in parameter list
    call message(AppMsg,Sparse,"Checking Input file...")
    !
    ! in: infile, xblockname, xcolumnname, nycols
    !out: xblockname,ycolumnsptr, ngoodcols,ninrows,goodcolsptr,in_set,in_tab
    !
    call check_table(infile, xblockname, xcolumnname, nycols, &
                      ninrows, ycolumnsptr, ngoodcols, goodcolsptr, &
                      in_set, in_tab)

    write(errstr,*)"Infile        = ",infile
    call message(AppMsg,Verbose)
    write(errstr,*)"xcolname      = ",xcolumnname
    call message(AppMsg,Verbose)
    write(errstr,*)"nycols        = ",nycols
    call message(AppMsg,Verbose)
    write(errstr,*)"points        = ",points
    call message(AppMsg,Verbose)
    write(errstr,*)"xblockname    = ",xblockname
    call message(AppMsg,Verbose)
    write(errstr,*)"ninrows       = ",ninrows
    call message(AppMsg,Verbose)
    write(errstr,*)"ngoodcols     = ",ngoodcols
    call message(AppMsg,Verbose)
    write(errstr,*)"gtiyes        = ",gtiyes
    call message(AppMsg,Verbose)

    do i = 1 , ngoodcols 
       write(errstr,*)"goodcols(",i,") = ",goodcolsptr(i)
       call message(AppMsg,Verbose)
    end do

    if (points <= 0) then
       write(errstr,*) &
            "Less than zero points per page, defaulting to nrows= ", &
            ninrows," in file: ",trim(infile)
       call warning("LessThanZeroPtPerPg")
       points = ninrows
    end if

    ! Get the y-column values. get_y_cols returns the 2D ptr array 
    ! ycolrowptr, which has each y-axis plot param in a column, and 
    ! all the values from the fits file for that column in it's rows 
    !
    ! in goes  : in_tab, goodcolsptr, ngoodcols, ninrows
    ! out comes: ycolunitsptr,ycolrowptr

    call get_y_cols(in_tab, goodcolsptr, ngoodcols, ninrows, &
                    ycolunitsptr, ycolrowptr)

    write(errstr, *) "Xcolumnname is: ",trim(xcolumnname)
    call message(AppMsg,Verbose)

    ! Now get the x-column from the fits file and fill a pointer array.
    ! x_out_ptr is 2 DIM, one for each ccd for FRAME plots, and
    ! the first DIM is simply 1 for non-FRAME plots. get_arrays_other
    ! will convert String-type TIME columns if necessary, using 
    ! OAL_ertToTimeTag.

    !################ FRAME-TYPE DATA ##################

    if (frames) then
       !
       ! only get convert parameter if x-axis is FRAME
       ! convertframe = convert FRAME column to real time
       ! (if used for x-axis)
       !
       if (convert) then
          timeyes = .TRUE.
          !          xcolumnname = "TIME"
          !          xcolunit = "s"
          !
          ! check for compatibilities between input params and CONVERTFRAME
          !
          if (.not. useodf) then
             call error("UseodfNoAndConvertYes", &
                  "Parameters useodf=no and convertframe=yes -- not compatable")
          end if
          if (.not. useccds) then
             call warning("UseCCDSnoAndConvertYes", &
                  "Parameters useccds=no and convertframe=yes -- may not be compatable")
          end if
       end if
       !
       ! fill x_out_ptr(ccds,ninrows) and ycolrowptr(ccds,ncols,ninrows)
       !
       call message(AppMsg,Sparse, &
            "FRAME type column for x-axis, sorting by ccd (if applicable)...")
       !
       ! get_arrays_frame returns x_out_ptr and y_out_ptr.
       ! these are from x_out_ptr(1:ccds,1:nrows)
       !            and y_out_ptr(1:ccds,1:ngoodcols,1:nrows)
       ! 
       call get_arrays_frame(in_tab, ninrows, ngoodcols , &
            xcolumnname, xcolunit, ycolrowptr , ccd_ptr, x_out_ptr, &
            y_out_ptr, n_ccd_row, xmin_ptr, xmax_ptr)
    elseif (other) then
       !################ OTHER-TYPE DATA (TIME, etc) ##################
       !
       ! fill x_out_ptr(1,ninrows) and ycolrowptr(1,ncols,ninrows)
       !
       if (index(trim(xcolumnname),trim(timetype)) /= 0) timeyes = .TRUE.
       call message(AppMsg,Sparse, &
            "OTHER type column for x-axis, sorting by ccd (if applicable)...")
       !
       ! get_arrays_other returns x_out_ptr and y_out_ptr.
       ! these are from x_out_ptr(1:ccds,1:nrows)
       !            and y_out_ptr(1:ccds,1:ngoodcols,1:nrows)
       ! 
       call message(AppMsg,Verbose,&
            "Reading FITS file with TIME or OTHER as x-axis...")
       !
       ! in goes  : in_tab,ninrows,ngoodcols,xcolumnname,ycolrowptr
       ! out comes: xcolunit,x_out_ptr,y_out_ptr,n_ccd_row
       !
       call get_arrays_other(in_tab, ninrows, ngoodcols , &
            xcolumnname, ycolrowptr, xcolunit, &
            x_out_ptr, y_out_ptr , n_ccd_row, xmin_ptr, xmax_ptr)

       write(errstr, *) "ninrows=",ninrows,"ngoodcols=",ngoodcols           
       call message(AppMsg,Noisy)
       write(errstr, *) "xcolumnname=",trim(xcolumnname),&
            "xcolunit=",trim(xcolunit)
       call message(AppMsg,Noisy)
    else 
       call error ("NotFRAMEorOTHER","Column is unknown type")
    end if
    !
    ! gtiyes=plot gti's at bottom of each plot eventually. This
    ! means extracting them...
    !
    if (timeyes) then 
       if (gtiyes) then                 ! set in getparams
          call message(AppMsg,Sparse, &
               "GTI's selected, will plot with each y-column.")
          call get_gti(gtifile, ccd_ptr, gti_ccd_ptr,&
               n_gti_row, maxgtirows)
          write (errstr,*) "allocating gti_array(",ccds,",",ninrows,")"
          call message (AppMsg,Sparse)
          allocate(gti_array(ccds,ninrows))
          
          gti_array = 0.0
          do ccdno = 1 , ccds
             if (aux) then
                write (errstr,*) "Converting GTIs for CCD: ",ccd_ptr(ccdno)
             else 
                write (errstr,*) "Creating GTI_ARRAY"
             end if

             call message(AppMsg,Verbose)
             
             do i = 1 , n_gti_row(ccdno)
                write (errstr,*) "CCD: ",ccd_ptr(ccdno), " Row: ",i,&
                     " Start:",gti_ccd_ptr(ccd_ptr(ccdno),1,i),&
                     " End: ",gti_ccd_ptr(ccd_ptr(ccdno),2,i)
                call message(AppMsg,Sparse)
             end do
             
             allocate(temp_gti(ninrows))
             allocate(temp_x(ninrows))
             temp_gti = 0.0
             temp_x(1:ninrows) = x_out_ptr(ccdno,1:ninrows)
             
             do i = 1 , n_gti_row(ccd_ptr(ccdno))
                ! for non-frame data, x_out_ptr will be (1,ninrows)
                ! because data is not by CCD.
                where ((temp_x >= gti_ccd_ptr(ccd_ptr(ccdno),1,i)) .AND. &
                     (temp_x <= gti_ccd_ptr(ccd_ptr(ccdno),2,i)))
                   temp_gti = 1.0
                end where
             end do
             
             do i = 1 , ninrows
                if (temp_gti(i) /= 0.0) then
                   gti_array(ccdno,i) = temp_gti(i)
                end if
             end do
             
             deallocate(temp_gti)
             deallocate(temp_x)
             
          end do
       end if
    end if

    allocate (minlox(2),minloy(3),maxlox(2),maxloy(3))    

    minlox = minloc(x_out_ptr)
    minloy = minloc(y_out_ptr)
    maxlox = maxloc(x_out_ptr)
    maxloy = maxloc(y_out_ptr)

!    write(errstr, *) "minlox=",minlox(1)," ",minlox(2)," maxlox=",maxlox(1)," ",maxlox(2)
!    call message(AppMsg,Sparse)
    
!    write(errstr, *) "minloy=",minloy(1)," ",minloy(2)," ",minloy(3)," maxloy=",maxloy(1)," ",maxloy(2)," ",maxloy(3)

!    call message(AppMsg,Sparse)
    
    if (getstartstop) then 
       tstartin = minval(xmin_ptr)
       tstopin = maxval(xmax_ptr)
       write(errstr,*) "Calculated TSTART,TSTOP = ",tstartin,tstopin
       call message(AppMsg,Noisy)
    else
       write(errstr,*) "Read from file TSTART,TSTOP ",tstartin,tstopin
       call message(AppMsg,Noisy)
    end if

    if (duration == 0.) then
       duration = tstopin - tstartin    ! bodge a value for plot label
    end if

    call message(AppMsg,Sparse,"Creating plot labels...")

    ! create labels for plot page

    call create_main_labels(infile, gtifile)

    ! correct for desired start and stop row if applicable

    if (gtiyes) then
       allocate (temp(1))
       temp = maxval(n_ccd_row)
       if (rown > temp(1)) then
          write(errstr,*) &
               "Lastrow selected > ninrows, defaulting to Max CCD row: ",temp(1)
          call warning("RowNGreaterThanNinrows")
          rown = temp(1)
       end if
       deallocate(temp)
    else
       if (rown > ninrows) then
          write(errstr,*) "Lastrow selected > ninrows, defaulting to ninrows"
          call warning("RowNGreaterThanNinrows")
       end if
    end if

    if (gtiyes) then 
!       allocate (gti_out(ccds,nrows))
!       gti_out(1:ccds,1:nrows) = gti_array(1:ccds,row1:rown)
!       deallocate(gti_array)
    end if

    ! call pgplot subroutines with collected data, plotting both
    ! columns selected and GTI intervals at the bottom of each page

    call message(AppMsg,Sparse,"Calling do_plot plotting subroutine")

    ! if offset requested, change bins from seconds since MJDREF
    ! to seconds since x_out(1) (ie. tstartin). Only for FRAME or
    ! TIME x-axis columns...

    if (offset) then
       call message(AppMsg,Sparse,"X-axis is offset to first value")
       do ccdno = 1 , ccds
          x_out_ptr(ccdno,1:n_ccd_row(ccdno)) = &
               x_out_ptr(ccdno,1:n_ccd_row(ccdno)) - tstartin
       end do
    end if

    do i = 1 , ccds
       do j = 1 , ninrows
          if (j < 6 .OR. j > (ninrows - 5)) then 
             write(errstr,*)"page_y(",i,",1,",j,")= ",y_out_ptr(i,1,j)
!             call message(AppMsg,Verbose)
          end if
       end do
    end do

    call do_plot(device, plotfile, xcolumnname, xcolunit, ycolumnsptr , &
         ycolunitsptr, points, offset, ngoodcols, ninrows, &
         n_ccd_row, y_out_ptr, x_out_ptr, ccd_ptr, gti_array)

    deallocate(x_out_ptr,y_out_ptr)

  end subroutine hkauxplot
      
  !------------------------------------------------------------------------
  subroutine getparams(infile, xblockname, xcolumnname, &
                       ycolnamesptr, gtifile, nycols, ccd_ptr, &
                       device, plotfile, points, offset) 
    !----------------------------------------------------------------------
    ! subroutine to retrieve paramaters from the hkauxplot.par file
    !
    ! infile       : Name of the input file (HK or AUX FITS file)
    ! xblockname   : Name of extension x-axis originates (typically 0)
    ! xcolumnname  : Name of x-axis column (usually TIME or FRAME)
    ! ycolnamesptr : pointer array of y-axis values to plot from xblock
    ! gtifile      : Name of the input gti file
    ! nycols       : Number of selected in parameters to plot
    ! device       : output plot device (e.g /ps)
    ! plotfile     : Name of output plot file (e.g foo.ps)
    ! points       : points per page to plot

    character(len=*), intent(out) :: infile, xblockname, xcolumnname, &
                                     gtifile,  plotfile, device
    character(len=*), dimension(:), pointer :: ycolnamesptr
    character(len=256) :: filename
    integer(kind=Int16),  dimension(:), pointer :: ccd_ptr, temp_ccd_ptr
    integer, intent(out) :: nycols
    integer :: np , points , i , tmp
    logical :: offset

    useodf = booleanParameter(useodfpar)

    ! get the x-axis column name to plot 
    ! default value for xcol is TIME
    
    xcolumnname = change_case(trim(stringParameter(xcolpar)),2)
    
    if (index(xcolumnname,trim(frametype)) /= 0) then
       frames = .TRUE.
       convert = booleanParameter(convertpar)
    else 
       other = .TRUE.
    end if

    ! get in file parameter. This can be a filename (e.g. 
    ! 0001_0000000000_R1X00100PEH.FIT or it's ODF convention
    ! ODF(RGS1,PERIODIC_HOUSEKEEPING).

    filename = trim(stringParameter(infilepar))

    ! if useodf=yes, the file is in ODF directory, regardless of format
    ! if useodf=no and x=frame and convertframe=yes, a fatal error occurs

    if (useodf) then
       write(errstr,*) "Reading input filename: ", trim(filename), " from SAS_ODFDIR"
       ! check filename format, expand if 'ODF(INST,MODE)'

       if ((index(trim(filename),"ODF(") == 0) .AND. &
            (index(trim(filename),"odf(") == 0)) then
          call message(AppMsg,Verbose,"Full input filename given.")
       else 
          call message(AppMsg,Verbose,&
               "Filename in ODF format given, expanding...")
       end if
       call message(AppMsg,Verbose,"Calling OAL...")
       call OAL_expandFileName(filename,infile)
       write(errstr,*) "expanded filename is: ",trim(infile)
       call message(AppMsg,Verbose)
       call message(AppMsg,Verbose,"Setting state in OAL...")
    else
       if (convert) then 
          call error ("CannotConvert","No FRAME convert to time without useodf set")
       end if
       infile = trim(filename)       
       write(errstr,*) "Reading input filename: ", trim(infile)
    endif

    call message(AppMsg,Verbose)
    
    ! get the x-axis block name to plot
    ! boolean "usexblk" determines whether to read
    ! the value for xblockname or default to 0 (1st extension)
       
    usexblk = booleanParameter(usexblkpar)        
    if (usexblk) then 
       xblockname = change_case(trim(stringParameter(xblkpar)),2)
    end if
    !
    ! get in column name parameters next
    !
    nycols = parameterCount(colnamespar)
    if (nycols <= 0) then
       call error ("NoYcolParams",&
            "No ycolumnnames selected")
    end if

    allocate(ycolnamesptr(nycols))    
    do np = 0, nycols-1
       ycolnamesptr(np+1) =  change_case(trim(stringParameter(colnamespar,np)),2)
    end do
    
    points = intParameter(pointspar)
    if (points > 999) then
       call warning("MoreThan1000PtPerPg",&
            "More than 1000 plot pts per page not recommended")
    end if
    
    offset = booleanParameter(offsetpar)
    selectrows = booleanParameter(selrowspar)
    selecttimes = booleanParameter(seltimepar)
    selectframes = booleanParameter(selframepar)
    
    if (selectrows) then
       if (selecttimes .OR. selectframes) then
          call error ("SelectsNotCompatible",&
               "Cannot select on more than one criteria: ROW|TIME|FRAME")
       end if
       start = realParameter(startRpar)
       stop = realParameter(stopRpar)
       call check_ss (start,stop,0)
    else
       getranges = .TRUE.
    end if
    
    if (selecttimes) then
       if (selectrows .OR. selectframes) then
          call error ("SelectsNotCompatible",&
               "Cannot select on more than one criteria: ROW|TIME|FRAME")
       end if
       start = realParameter(startTpar)
       stop = realParameter(stopTpar)
       call check_ss (start,stop,1)
    else
       getranges = .TRUE.
    end if
    
    if (selectframes) then
       if (selectrows .OR. selecttimes) then
          call error ("SelectsNotCompatible",&
               "Cannot select on more than one criteria: ROW|TIME|FRAME")
       end if
       start = realParameter(startFpar)
       stop = realParameter(stopFpar)
       call check_ss (start,stop,2)
    else
       getranges = .TRUE.
    end if

    ! now get possible gti parameter information
    
    gtiyes = booleanParameter(usegtipar)
    
    if (gtiyes) then
       gtifile = trim(stringParameter(gtisetpar))
       if (len_trim(gtifile) <= 0) then
          call error("noGTIfile","Blank input GTI file name selected.")
       else 
          write(errstr,*) "Input GTI file : ",trim(gtifile)
          call message(AppMsg,Sparse)
       end if
    else
       write(errstr,*) "A GTI file was not selected. GTIs will not be plotted"
       call message(AppMsg,Sparse)
       gtiyes = .FALSE.
    end if
    
    call getnccds(trim(infile))  ! returns nccds = total available ccds
    
    useccds = booleanParameter(useccdspar)
    i = 0
    selectccds = .false.         ! don't select ccds as default
    
    if (useccds) then            ! plot ANY ccds?
       selectccds = booleanParameter(selccdpar)
       if (selectccds) then      ! select individual ccds?
          allocate (temp_ccd_ptr(nccds))
          ccds = parameterCount(trim(ccdpar))
          do np = 0, ccds-1
             tmp = intParameter(ccdpar,np)
             if (tmp <= nccds) then
                i = i + 1
                temp_ccd_ptr(i) =  intParameter(ccdpar,np)
             else 
                write(errstr,*) "CCD selected: ",tmp,&
                     " not in instrument range -- not used"
                call warning("CCDselectedInvalid")
             end if
          end do
          if (i == 0) then 
             write(errstr,*) "None of CCDs selected are Valid!"
             call error("AllCCDselectedInvalid")
          else
             !
             ! an array of the ordered ccd's selected
             ! eg 1 3 2 7 5
             !
             allocate(ccd_ptr(i))
             ccd_ptr(1:i) = temp_ccd_ptr(1:i)
             deallocate (temp_ccd_ptr)
             ccds = i    ! change ccds to reflect "good" ccds
          end if
       else 
          call message(AppMsg,Verbose,"Using all ccds in infile.")
          allocate(ccd_ptr(nccds))
          do np = 1 , nccds
             ccd_ptr(np) = np
          end do
          ccds = nccds
       end if
    else 
       gtiyes = .false.    ! override gtiyes if not CCDs
    end if
           
    device = change_case(StringParameter(devicepar),2)
    selectoutfile = booleanParameter(selectoutpar)
    if (selectoutfile) then
       plotfile = trim(stringParameter(outfilepar))
       write(errstr,*) "Output file : ",trim(plotfile)
       call message(AppMsg,Sparse)
    else
       call message(AppMsg,Verbose,&
            "Output file not selected, plotting to /XW")
    end if
    
  end subroutine getparams
  !------------------------------------------------------------------------
  subroutine getnccds(infile)
    !----------------------------------------------------------------------

    character(len=*), intent(in)  :: infile
    type(DataSetT)  :: in_set
    type(TableT)    :: in_tab

    write(errstr,*) "Checking for CCDs in file: ",infile
    call message(AppMsg,Verbose)

    write(errstr,*) "Opening for read, file: ",infile
    call message(AppMsg,Verbose)

    in_set = dataSet(trim(infile), Read)

    call message(AppMsg,Verbose,"Reading primary header...")

    in_tab = table(in_set, 0)

    if (hasAttribute(in_set,'TELESCOP')) then
       telescop = trim(stringAttribute(in_set,'TELESCOP'))
    elseif(hasAttribute(in_tab,'TELESCOP')) then
       telescop = trim(stringAttribute(in_tab,'TELESCOP'))
    else
       telescop = "XMM"
       call warning("noTELESCOP","TELESCOP attribute missing, using XMM.")
    end if
    !
    ! chose mos1 because it has the least ccds of mos/rgs
    !
    if (hasAttribute(in_set,'INSTRUME')) then
       instrume = trim(stringAttribute(in_set,'INSTRUME'))
    elseif(hasAttribute(in_tab,'INSTRUME')) then
       instrume = trim(stringAttribute(in_tab,'INSTRUME'))
    else
       instrume = "NONE"
       call warning("noINSTRUME","INSTRUME attribute missing, using NONE.")
    end if

    if (hasAttribute(in_set,'OBS_ID')) then
       obs_id = trim(stringAttribute(in_set,'OBS_ID'))
    elseif(hasAttribute(in_tab,'OBS_ID')) then
       obs_id = trim(stringAttribute(in_tab,'OBS_ID'))
    else
       obs_id = "00000"
       call warning("noOBS_ID","OBS_ID attribute missing, using 00000")
    end if

    if (hasAttribute(in_set,'EXP_ID')) then
       exp_id = trim(stringAttribute(in_set,'EXP_ID'))
    elseif(hasAttribute(in_tab,'EXP_ID')) then
       exp_id = trim(stringAttribute(in_tab,'EXP_ID'))
    else
       exp_id = "00000"
       call warning("noEXP_ID","EXP_ID attribute missing, using 00000")
    end if

    if (hasAttribute(in_set,'DURATION')) then
       duration = real32Attribute(in_set,'DURATION')
    elseif(hasAttribute(in_tab,'DURATION')) then
       duration = real32Attribute(in_tab,'DURATION')
    else
       duration = 0.0
       call warning("noDURATION","DURATION attribute missing, using 0.0")
    end if

    if (hasAttribute(in_set,'TSTART')) then
       tstartin = real32Attribute(in_set,'TSTART')
    elseif(hasAttribute(in_tab,'TSTART')) then
       tstartin = real32Attribute(in_tab,'TSTART')
    else
       tstartin = -999.0
       getstartstop = .true.
       call warning("noTSTART","TSTART attribute missing, will calculate.")
    end if

    if (hasAttribute(in_set,'TSTOP')) then
       tstopin = real32Attribute(in_set,'TSTOP')
    elseif(hasAttribute(in_tab,'TSTOP')) then
       tstopin = real32Attribute(in_tab,'TSTOP')
    else
       tstopin = -999.0
       call warning("noTSTOP","TSTOP attribute missing, will calculate.")
    end if

    if (hasAttribute(in_set,'MJDREF')) then
       mjdref = real32Attribute(in_set,'MJDREF')
    elseif(hasAttribute(in_tab,'MJDREF')) then
       mjdref = real32Attribute(in_tab,'MJDREF')
    else
       mjdref = 50812.5972222222
       call warning("noMJDREF","MJDREF attribute missing, using 50812.6.")
    end if

    if (hasAttribute(in_set,'DATE-OBS')) then
       date_obs = trim(stringAttribute(in_set,'DATE-OBS'))
    elseif(hasAttribute(in_tab,'DATE-OBS')) then
       date_obs = trim(stringAttribute(in_tab,'DATE-OBS'))
    else
       date_obs = "1999-12-15T00:00:00"
       call warning("noDATE-OBS","DATE-OBS attribute missing, using 1999-12-15T12:00:00.")
    end if

    if (hasAttribute(in_set,'DATE-END')) then
       date_end = trim(stringAttribute(in_set,'DATE-END'))
    elseif(hasAttribute(in_tab,'DATE-END')) then
       date_end = trim(stringAttribute(in_tab,'DATE-END'))
    else
       date_end = "1999-12-15T00:00:00"
       call warning("noDATE-END","DATE-END attribute missing, using 1999-12-15T12:00:00")
    end if

    if (hasAttribute(in_set,'OBJECT')) then
       object = trim(stringAttribute(in_set,'OBJECT'))
    elseif(hasAttribute(in_tab,'OBJECT')) then
       object = trim(stringAttribute(in_tab,'OBJECT'))
    else
       object = "No object listed"
       call warning("noOBJECT"," attribute missing, using: No object listed")
    end if


    call release(in_tab)
    call release(in_set)

    ! check for the number of ccds

    select case (trim(instrume))
    case ("EMOS1")
       nccds = 7
    case ("EMOS2")
       nccds = 7
    case ("EPN")
       nccds = 12
    case ("RGS1")
       nccds = 9
    case ("RGS2")
       nccds = 9
    case DEFAULT 
       nccds = 1
    end select
    
    write(errstr,*)"Instrument: ",trim(instrume),"  has ",nccds," ccds"
    call message(AppMsg,Verbose)

  end subroutine getnccds

  !------------------------------------------------------------------------
  subroutine check_table(infile,xblockname,xcolumnname,nycols,&
                         ninrows,ycolumnsptr,ngoodcols,goodcolsptr,&
                         in_set,in_tab)
    !----------------------------------------------------------------------
    ! subroutine to open, check and read the in file block (extension)
    !
    ! infile : name of input IN FITS file
    ! xblockname : name of IN FITS file extension containing columns requested
    ! xcolumnname : name of column to be used for plot x-axis
    ! nycols : number of y columns to be plotted
    ! ycolsptr : pointer array of columns requested (size nycols)
    ! ngoodcols : number of columns requested actually found in infile
    ! goodcolsptr : new pointer array of columns found and ok to plot
    ! in_set : in DAL file set
    ! in_tab : in DAL table

    character(len=*), intent(in)  :: infile ,  xcolumnname
    character(len=*), intent(inout) :: xblockname
    character(len=80) :: datatype , expression
    character(len=80) , dimension(:), allocatable :: out_cols , col_unit
    character(len=*), dimension(:), pointer :: ycolumnsptr , goodcolsptr
!    character(len=19), allocatable, dimension(:) :: temp_string

    integer :: ngoodcols , nycols , i , coltype , ninrows , nrow1
    integer, dimension(:), allocatable :: col_type
    type(DataSetT), intent(out)  :: in_set
    type(DataSetT)  :: temp_set
    type(TableT),   intent(out)  :: in_tab
    type(TableT) :: temp_tab
    type(ColumnT)  :: in_col , out_col
    
    ! open in file and read in extension table (block)

    in_set = dataSet(infile, Read)

    ! if set usexblk in getparameters, open requested block

    if (usexblk) then 
       in_tab = table(in_set, trim(xblockname))
    else 
       in_tab = table(in_set, 0)        ! default to first extension
       xblockname = trim(name(in_tab))  ! return 1st extension name
    end if

    write(errstr,*) "Reading input file extension ",trim(xblockname)
    call message(AppMsg,Verbose)

    if(.not. hasAttribute(in_tab,dtpar)) then
       write(errstr,*) "Input file extension does not contain DATATYPE attribute"
       call warning("NoDatatypeAttrib")
       other = .TRUE.
    else
       datatype = trim(stringAttribute(in_tab,dtpar))
       if (index(datatype,trim(hkdatatype)) /= 0) then
          hk = .TRUE.
       elseif (index(datatype,trim(auxdatatype)) /= 0) then
          aux = .TRUE.
       else
          other = .TRUE.
       end if
    end if

    write (errstr,*) "Datatype is: ",trim(datatype)
    call message(AppMsg,Verbose)

    ! get columns to plot from parameter table

    write(errstr,*) "Checking for parameters in table..."
    call message(AppMsg,Sparse)

    if (.not. hasColumn(in_tab,trim(xcolumnname))) then
       write(errstr,*) "x-axis column: ",trim(xcolumnname)," not in infile "
       call error ('XcolNotFound')
    end if

    ! Loop through selected columns. If in column not found,
    ! warn but plot any that ARE present. If in column a string,
    ! warn and deselect, but plot any others
    
    allocate(out_cols(nycols),col_type(nycols),col_unit(nycols))
    ngoodcols = 0
    !
    ! set up a temp set in memory with only the desired columns
    ! [especially helpful for PEH data with 1000+ columns but 
    ! only a few needed for plotting]
    !
    ninrows = numberOfRows(in_tab)
    write(errstr,*) "numrows = ",ninrows
    call message(AppMsg,Verbose)
    !
    ! add the x-axis column first
    !
    in_col = column(in_tab, trim(xcolumnname), Read)
    temp_set = dataSet('hkaux.temp', Temp)
    temp_tab = addTable(temp_set,trim(xblockname),ninrows)

    call addAttributes(attributable(temp_set),attributable(in_set))
    call addAttributes(attributable(temp_tab),attributable(in_tab))

    call copyColumn(temp_tab, in_col)
    out_col = column(temp_tab, xcolumnname, modify)

    call addAttributes(attributable(out_col),attributable(in_col))
    call release(out_col)
    call release(in_col)
    !
    ! add the CCDID column (if applicable) need it later if 
    ! sorting or using FRAME data
    !
    if (hasColumn(in_tab,trim(ccdcol))) then
       write(errstr,*) "Writing CCDID column "
       call message(AppMsg,Verbose)
       in_col = column(in_tab, trim(ccdcol), Read)
       call copyColumn(temp_tab, in_col)
       out_col = column(temp_tab, ccdcol, modify)
       call addAttributes(attributable(out_col),attributable(in_col))
    end if
    call release(out_col)
    call release(in_col)
    !
    ! add the FRAME column (if applicable) need it later when 
    ! sorting by CCDID
    !
    if (hasColumn(in_tab,trim(frametype))) then
       if (index(xcolumnname,trim(frametype)) == 0) then
          write(errstr,*) "Writing FRAME column "
          call message(AppMsg,Verbose)
          in_col = column(in_tab, trim(ccdcol), Read)
          call copyColumn(temp_tab, in_col)
          out_col = column(temp_tab, ccdcol, modify)
          call addAttributes(attributable(out_col),attributable(in_col))
       end if
    end if
    call release(out_col)
    call release(in_col)

    do i = 1, nycols

       if (.not. hasColumn(in_tab,trim(ycolumnsptr(i)))) then
          write(errstr,*) "column ",trim(ycolumnsptr(i))," not in input file "
          call warning ('YcolNotFound')
       else
          !
          ! found column so put into tempfile
          !
          in_col = column(in_tab, trim(ycolumnsptr(i)), Read)
          coltype = columnDataType(in_col)
          write(errstr, *) "column is ",trim(ycolumnsptr(i))," Type is ",coltype
          call message(AppMsg,Verbose)
          select case (coltype)

          case(String)
             write(errstr, *) "String column Type not plottable"
             call warning("StringColNotPlottable")
             cycle
          case DEFAULT
             select case (coltype)
                !       case(Boolean)
                !          temp_bool => booleanData(in_col)
                !          temp_real = real(temp_bool) 
             case(UInteger32)
                                call message(AppMsg,Verbose,"Unsigned integer, will return signed data!")
             case(UInteger16)
                                call message(AppMsg,Verbose,"Unsigned integer, will return signed data!")
             case(Integer32)
                !                call message(AppMsg,Verbose,"I32")
             case(Integer16)
                !                call message(AppMsg,Verbose,"I16")
             case(Integer8)
                !                call message(AppMsg,Verbose,"I8")
             case(Real32)
                !                call message(AppMsg,Verbose,"R32")
             case(Real64)
                !                call message(AppMsg,Verbose,"R64")
             case DEFAULT
                write(errstr, *) "Column Type: ",coltype," not plottable"
                call warning("colNotPlottable")
                cycle
             end select
             if (.not. hasColumn(temp_tab,trim(ycolumnsptr(i)))) then
                write(errstr, *) "column ok, adding ",trim(ycolumnsptr(i))," to temp_set"
                call message(AppMsg,Verbose)
                ngoodcols = ngoodcols + 1
                out_cols(ngoodcols) = ycolumnsptr(i)
                col_type(ngoodcols) = coltype
                col_unit(ngoodcols) = Units(in_col)
                call copyColumn(temp_tab, in_col)
                out_col = column(temp_tab, out_cols(ngoodcols) , modify)
                call addAttributes(attributable(out_col),attributable(in_col))
             else 
                write(errstr, *) "NOT adding ",trim(ycolumnsptr(i))," to temp_set  (already exists)"
                call message(AppMsg,Verbose)
             end if
          end select
          call release(out_col)
          call release(in_col)
       end if
    end do

    call release (in_set)

    in_set = temp_set
    in_tab = table(in_set, 0)
    !
    ! filter table if selectrows or selectframes chosen (must wait
    ! until Frame converted to Time for selecttimes)
    !
    ! Make the expression for initFilter for FRAME
    !
    if (selectframes) then
       write(expression,'("(FRAME>=",I4," && FRAME<=",I4,")")') int(start),int(stop)
       write(errstr,*) "SelectFRAME initFilter with expression ", expression
       call message(AppMsg,Verbose)
       call initFilter( expression )       
       nrow1 = doFilter( in_tab )      ! Filter table
       if (nrow1 < 1) then
          write(errstr,*) "The input file ", trim(infile),  &
               " does not contain any data for FRAMEs ", int(start)," to ",int(stop)
          call error("FRAMErangeNotValid")
       endif
    elseif (selectrows) then
       write(expression,'("(#ROW>=",I9," && #ROW<=",I9,")")') int(start),int(stop)
       write(errstr,*) "SelectROW initFilter with expression ", expression
       call message(AppMsg,Verbose)
       call initFilter( expression )       
       nrow1 = doFilter( in_tab )      ! Filter table
       if (nrow1 < 1) then
          write(errstr,*) "The input file ", trim(infile),  &
               " does not contain any data for FRAMEs ", int(start)," to ",int(stop)
          call error("ROWrangeNotValid")
       endif
    end if

    if (ngoodcols <= 0 ) then
       write(errstr,*) "No good Y columns found in file ",trim(infile)
       call error ('NoYcolsToPlot')
    end if

    ! more than 20 columns per page looks bad!

    if (ngoodcols > 20) then
       write(errstr,*) "More than 20 plot columns not recommended"
       call warning("MoreThan20Cols",&
            "More than 20 plot columns not recommended")
    end if

    if (ngoodcols < nycols ) then
       write(errstr,*) "Only ",ngoodcols,&
            " of the columns entered were found in input file."
       call warning ('notAllcolsPlot')
    end if
    !
    ! fix on selectframe and selectrows
    !
    ninrows = numberOfRows(in_tab)
    write(errstr,*) "numrows = ",ninrows
    call message(AppMsg,Verbose)

    ! Copy required for selectlib

    allocate (goodcolsptr(1:ngoodcols))
    goodcolsptr(1:ngoodcols) = out_cols(:ngoodcols)
    deallocate(out_cols)

  end subroutine check_table
  
  !----------------------------------------------------------------------
  subroutine get_y_cols(in_tab, goodcolsptr, ngoodcols, ninrows, &
                        ycolunitsptr, ycolrowptr)
    !--------------------------------------------------------------------
    ! in:  in_tab, goodcolsptr, ngoodcols, ninrows
    ! out: ycolunitsptr, ycolrowptr
    !
    character(len=80), dimension(:), pointer :: goodcolsptr, ycolunitsptr
    integer, intent(in) :: ngoodcols , ninrows
    integer :: icol , ycoltype , i
    real(single), allocatable, dimension(:) :: temp_real
    integer(kind=Int32),  dimension(:), pointer :: temp_int32
    integer(kind=Int16),  dimension(:), pointer :: temp_int16
    integer(kind=Int8),   dimension(:), pointer :: temp_int8
    real(kind=double),    dimension(:), pointer :: temp_real64
    real(kind=single),    dimension(:), pointer :: temp_real32
    real(kind=double), dimension(:,:), pointer :: ycolrowptr

    type(TableT), intent(in)   :: in_tab
    type(ColumnT) :: in_col

    write(errstr,*)"Allocating ycolrowptr(",ngoodcols,",",ninrows,")"
    call message(AppMsg,Verbose)

    allocate (ycolrowptr(1:ngoodcols,1:ninrows))
    allocate (ycolunitsptr(1:ngoodcols))

    call message(AppMsg,Verbose,"Reading column data for y-axis...")

    do icol = 1, ngoodcols

       allocate (temp_real(ninrows))
       in_col = column(in_tab, goodcolsptr(icol), read)
       ycoltype = columnDataType(in_col)
       ycolunitsptr(icol) = Units(in_col)
       
       select case (ycoltype)
          !       case(Boolean)
          !          temp_bool => booleanData(in_col)
          !          temp_real = real(temp_bool) 
       case(Integer32)
          temp_int32 => int32Data(in_col)
          temp_real = real(temp_int32) 
       case(Integer16)
          temp_int16 => int16Data(in_col)
          temp_real = real(temp_int16)
       case(Integer8)
          temp_int8 => int8Data(in_col)
          temp_real = real(temp_int8)
       case(Real32)
          temp_real32 => real32Data(in_col)
          temp_real = real(temp_real32)
       case(Real64)
          temp_real64 => real64Data(in_col)
          temp_real = real(temp_real64)   ! truncates, but too bad...
       case DEFAULT
          write(errstr, *) "Column Type not supportable"
          call error("TypeError")
       end select
       !
       ! fill the cols array with each column of data to plot
       !
       ycolrowptr(icol,1:ninrows) = temp_real(1:ninrows)
       do i = 1 , 10
          write (errstr,*) "ycolrowptr(",icol,",",i,") = ",ycolrowptr(icol,i)
          call message(AppMsg,Verbose)
       end do
       deallocate(temp_real)
       call release(in_col)
    end do

  end subroutine get_y_cols

  !-------------------------------------------------------------------
  subroutine get_arrays_frame(in_tab, ninrows, ngoodcols, &
                             xcolumnname, xcolunit, ycolrowptr, &
                             ccd_ptr, x_out_ptr, y_out_ptr, &
                             n_ccd_row, xmin_ptr, xmax_ptr)
    !-----------------------------------------------------------------

    character(len=80), intent(out) :: xcolunit
    character(len=*), intent(in) :: xcolumnname
    integer :: inst, dmode, expnr, ccdnr, nodenr, i, j, ngood
    integer, intent(inout) :: ninrows , ngoodcols 
    integer :: xcoltype , nrow , ccdno , coltype , col , goodrows , bigrows
    logical :: attfrom
    integer, dimension(:), allocatable :: ntimes
    integer(kind=Int32),  dimension(:), pointer :: frame_ptr, ccdid_ptr,&
                                                   n_ccd_row , tempx_ptr,&
                                                   ncr , vect_ccd, g_ccd_row
    integer(kind=Int32),  dimension(:,:), pointer :: allx_ptr
    !    integer(kind=Int32),  dimension(:), pointer :: temp_ptr
    integer(kind=Int16),  dimension(:), pointer :: temp_int16 , ccd_ptr,&
                                                   g_ccd_ptr
    integer(kind=Int8),   dimension(:), pointer :: temp_int8
    !    logical(kind=bool),   dimension(:), pointer :: temp_bool
    real(kind=double) :: xmin, xmax 
    real(kind=double),  dimension(:), pointer :: obtout, tempy_ptr, y_ptr, &
                                                 vect_y, xmin_ptr, xmax_ptr,&
                                                 timeptr
    real(kind=double),  dimension(:,:), pointer :: x_out_ptr, x_full_ptr, &
                                                   ycolrowptr
    real(kind=double),  dimension(:,:,:), pointer :: y_out_ptr , ally_ptr , &
                                                     y_full_ptr
    type(TableT)  :: in_tab
    type(ColumnT) :: in_col , in_col2

    in_col = column(in_tab,trim(xcolumnname), READ)
    xcoltype = columnDataType(in_col)
    xcolunit = Units(in_col)

    ! frame_ptr stores the column FRAME
    allocate (frame_ptr(0:ninrows-1))
    !
    ! FRAME column MUST be an integer, but write to 32-bit integer
    ! column to pass to OAL later if applicable.
    !
    select case (xcoltype)
    case(Integer32)
       frame_ptr => int32Data(in_col)
    case(Integer16)
       temp_int16 => int16Data(in_col)
       frame_ptr = temp_int16
    case(Integer8)
       temp_int8 => int8Data(in_col)
       frame_ptr = temp_int8
    case DEFAULT
       call error("FrameTypeNotInt", &
            "FRAME column data type must be Integer.")
    end select

    call release(in_col)

    ! if CCDID column exists, must get and sort x column by CCD

    if (hasColumn(in_tab,trim(ccdcol))) then

       call message(AppMsg,Verbose,"File has multiple CCDs...")
       
       if (.not. useccds) then
          call error ("useccdsNotSelected",&
               "Infile has by-CCD data, but useccds parameter not selected")
       end if

       write(errstr,*)"allocating ccdid_ptr with ",ninrows," rows"
       call message(AppMsg,Verbose)

       allocate (ccdid_ptr(0:ninrows-1))

       write(errstr,*)"getting column ",trim(ccdcol)
       call message(AppMsg,Verbose)
       !
       ! get the CCDID column. It can be any integer, but 
       ! write to 32-bit integer pointer array
       !
       in_col2 = column(in_tab,trim(ccdcol), READ)
       coltype = columnDataType(in_col2)
       select case (coltype)
       case(Integer32)
          ccdid_ptr => int32Data(in_col2)
       case(Integer16)
          temp_int16 => int16Data(in_col2)
          ccdid_ptr = temp_int16
       case(Integer8)
          temp_int8 => int8Data(in_col2)
          ccdid_ptr = temp_int8
       case DEFAULT
          call error("CCDIDTypeNotInt",&
               "CCDID column data type should be Integer.")
       end select
       call release(in_col2)
       !
       ! get nrows for each selected ccd
       ! 
       ! n_ccd_row is an array the collects the number of rows in
       ! related to a given ccd. ie. it correlates which row in a
       ! column goes with which ccd.

       allocate (n_ccd_row(ccds),ncr(ccds),g_ccd_row(ccds),g_ccd_ptr(ccds))
       n_ccd_row = 0
       g_ccd_row = 0
       ncr = 0
       j = 0
       !
       ! COUNT number of rows each ccd contributes to file
       !
       do i = 1 , ccds
          n_ccd_row(i) = count( ccdid_ptr == (ccd_ptr(i)) )
          write(errstr,*)"number of frames where CCDID = ",&
                          ccd_ptr(i)," is: ",n_ccd_row(i)
          call message(AppMsg,Verbose)          
          if (n_ccd_row(i) == 0) then
             write(errstr,*)"no data will be plotted for CCDID = ",ccd_ptr(i)
             call warning("CCDselectedInvalid")
          else
             j = j + 1
             g_ccd_row(j) = n_ccd_row(i)
             g_ccd_ptr(j) = ccd_ptr(i)
          end if
       end do
       if (j == 0) then 
          call error ("NoCCDdata","No FRAMES were found for ANY selected CCD")
       end if
       ccds = j
       deallocate(n_ccd_row,ccd_ptr)
       allocate(n_ccd_row(j),ccd_ptr(j))
       n_ccd_row(1:j) = g_ccd_row(1:j)
       ccd_ptr(1:j) = g_ccd_ptr(1:j)
       deallocate(g_ccd_row,g_ccd_ptr)
       goodrows = maxval(n_ccd_row)
       write(errstr,*)"max val of n_ccd_row is: ",goodrows
       call message(AppMsg,Verbose)
       
       write(errstr,*)"allocating ally_ptr(",ccds,",",ngoodcols,",",goodrows,")"
       call message(AppMsg,Verbose)
       !
       ! allocate y axis array to fill with y-axis frame data
       !
       allocate (ally_ptr(ccds,1:ngoodcols,goodrows))
       ally_ptr = 0.0     ! initialize y pointer array
       !
       ! allocate temp x-axis array
       !
       write(errstr,*)"allocating allx_ptr(",ccds,",",goodrows,")"
       call message(AppMsg,Verbose)
       allocate (allx_ptr(ccds,goodrows))
       !
       ! loop through selected ccds and fill x,y arrays
       ! ccds variable is number of ccds selected on command line

       allocate (tempy_ptr(ninrows))

       do i = 1 , ccds
          write(errstr,*)"allocating tempx_ptr(",n_ccd_row(i),")"
          call message(AppMsg,Verbose)
          write(errstr,*)"allocating pack vect_ccd(",n_ccd_row(i),")"
          call message(AppMsg,Verbose)
          allocate (tempx_ptr(n_ccd_row(i)),vect_ccd(n_ccd_row(i)))
          vect_ccd = 0   ! zero out pack vector
          tempx_ptr = 0  ! zero out pack array

          write(errstr,*)"packing tempx_ptr with frame: ",ccd_ptr(i)," data"
          call message(AppMsg,Verbose)
          tempx_ptr = pack (frame_ptr,MASK = ccdid_ptr == (ccd_ptr(i)), VECTOR = vect_ccd)
          !
          ! fill the allx_ptr array by each ccd with the current
          ! values in tempx_ptr, allx_ptr will be fed to oal...
          !
          allx_ptr(i,1:n_ccd_row(i)) = tempx_ptr(1:n_ccd_row(i))
          deallocate (vect_ccd)
       end do
       !
       ! do the y cols now
       !
       do j = 1 , ngoodcols
          do i = 1 , ccds
             allocate (y_ptr(n_ccd_row(i)),vect_y(n_ccd_row(i)))
             tempy_ptr = ycolrowptr(j,1:ninrows)
             y_ptr = pack (tempy_ptr,MASK = ccdid_ptr == (ccd_ptr(i)),VECTOR = vect_y)
             ally_ptr(i,j,1:n_ccd_row(i)) = y_ptr(1:n_ccd_row(i))
             deallocate (y_ptr,vect_y)
          end do
       end do

       deallocate (tempx_ptr)
    else
       ccds = 1  ! only one ccd if no CCDID in file

       ! n_ccd_row is an array the collects the number of rows
       ! related to a given ccd. ie. it correlates which row in a
       ! column goes with which ccd.

       n_ccd_row(1) = ninrows
       goodrows = ninrows

       write(errstr,*)"allocating allx_ptr(",ccds,",",ninrows,")"
       call message(AppMsg,Verbose)
       allocate (allx_ptr(ccds,ninrows))

       write(errstr,*)"allocating ally_ptr(",ccds,",",ngoodcols,",",ninrows,")"
       call message(AppMsg,Verbose)
       allocate (ally_ptr(1,1:ngoodcols,1:ninrows))
       allocate (n_ccd_row(1))
       do nrow = 0 , ninrows - 1
          allx_ptr(1,nrow) = frame_ptr(nrow) 
          do col = 1 , ngoodcols
             ally_ptr(1,col,nrow+1) = ycolrowptr(col,nrow+1)
          end do
       end do
    end if          

    ! take the input FRAME column and convert it to TIME
    ! by calling the OAL conversion subroutines (if the 
    ! user requests that by setting convert=yes, otherwise,
    ! task will simply plot FRAME number as x-axis.
    ! This task will ONLY convert a FRAME column to TimeTag
    ! if useodf is set (otherwise, OAL doesn't know where 
    ! to get the time conversions!

    !
    ! got to here, have allx_ptr and ally_ptr. All the data
    ! for the two arrays
    ! allx_ptr will be fed to OAL to convert to time, otherwise
    ! fill x_out_ptr and y_out_ptr respectively

    write(errstr,*)"allocating y_full_ptr(",ccds,",",ngoodcols,",",goodrows,")"
    call message(AppMsg,Verbose)
    allocate (y_full_ptr(ccds,1:ngoodcols,1:goodrows))
    write(errstr,*)"allocating x_full_ptr(",ccds,",",goodrows,")"
    call message(AppMsg,Verbose)
    allocate (x_full_ptr(ccds,goodrows))
    allocate (xmin_ptr(ccds),xmax_ptr(ccds))

    if (convert) then    ! if converting FRAME to real TIME

       call message(AppMsg,Sparse, &
            "Converting FRAME column to real TIME...")

       do ccdno = 1 , ccds

          write(errstr,*)"Calling OAL for ccdno: ",ccd_ptr(ccdno)
          call message(AppMsg,Sparse)

          call OAL_getState(instrument = inst , dataMode = dmode ,&
               exposureNr = expnr , ccdNr = ccdnr , nodeNr = nodenr ,&
               attitudeFromAhf = attfrom )

!          write(errstr,*)"inst=",inst,"dmode=",dmode,"expnr=",expnr,&
!               "ccdnr=",ccdnr,"nodenr=",nodenr,"attfrom=",attfrom
!          call message(AppMsg,Sparse)

          call OAL_setState(instrument = inst, dataMode = dmode , &
               exposureNr = expnr , ccdNr = INT(ccd_ptr(ccdno)))

          call OAL_getState(instrument = inst , dataMode = dmode ,&
               exposureNr = expnr , ccdNr = ccdnr , nodeNr = nodenr ,&
               attitudeFromAhf = attfrom )

          write(errstr,*)"inst=",inst,"dmode=",dmode,"expnr=",expnr,&
               "ccdnr=",ccdnr,"nodenr=",nodenr,"attfrom=",attfrom
          call message(AppMsg,Sparse)
          write(errstr, *) "Converting FRAMES of CCD:",ccd_ptr(ccdno),&
               "to Obt"
          call message(AppMsg,Sparse)

          allocate (tempx_ptr(1:n_ccd_row(ccdno)))
          tempx_ptr(1:n_ccd_row(ccdno)) = allx_ptr(ccdno,1:n_ccd_row(ccdno))
          call OAL_frameCounterToObt(tempx_ptr,obtout)
          deallocate (tempx_ptr)

          do i = 1 , 10 
             write(errstr, '(a,i3.3,a,i6)') "obtout(",i,")= ",NINT(obtout(i))
             call message(AppMsg,Sparse)
          end do

          write(errstr, *) "ccdno,ccd_ptr(ccdno),n_ccd_row(ccdno)=",&
               ccdno,ccd_ptr(ccdno),n_ccd_row(ccdno)
          call message(AppMsg,Sparse)

          write(errstr, *) "Converting OBT of CCD:",ccd_ptr(ccdno),&
               "to TimeTag"
          call message(AppMsg,Verbose)

          call OAL_obtToTimeTag(obtout,timeptr)

          do i = 1 , 10 
             write(errstr, *) "timeptr(",i,")= ",timeptr(i)
             call message(AppMsg,Verbose)
          end do

          ngood = count (obtout /= -1.0)

          write(errstr, *) "ngood = ",ngood
          call message(AppMsg,Sparse)

!          x_full_ptr(ccdno,1:n_ccd_row(ccdno)) = timeptr(0:n_ccd_row(ccdno)-1)
          j = 0
          xmin = HUGE(1.0)
          xmax = -(HUGE(1.0))
          do i = 1 , n_ccd_row(ccdno)
             if (obtout(i-1) /= -1.0) then
                j = j + 1
                x_full_ptr(ccdno,j) = timeptr(i-1)
                if (timeptr(i-1) < xmin) xmin = timeptr(i-1)
                if (timeptr(i-1) > xmax) xmax = timeptr(i-1)
                do col = 1, ngoodcols
                   y_full_ptr(ccdno,col,j) = ally_ptr(ccdno,col,i)
                end do
                !                write(errstr, *) "x_full_ptr(",ccdno,",",j,") = ",x_full_ptr(ccdno,j)
                !                call message(AppMsg,Sparse)
             end if
          end do
          xmin_ptr(ccdno) = xmin
          xmax_ptr(ccdno) = xmax
          write(errstr,*)"OLD n_ccd_row(",ccdno,")= ",goodrows
          call message(AppMsg,Sparse)
          n_ccd_row(ccdno) = ngood
          write(errstr,*)"NEW n_ccd_row(",ccdno,")= ",ngood
          call message(AppMsg,Sparse)
          write(errstr,*)"xmin,xmax of ",ccdno," = ",xmin,xmax
          call message(AppMsg,Sparse)
       end do
       !
       ! times converted, so select on times
       !
       bigrows = maxval(n_ccd_row)
       allocate (x_out_ptr(ccds,bigrows),y_out_ptr(ccds,col,bigrows))
!       x_out_ptr = HUGE(1.0)   ! put extra points off the graph
       x_out_ptr = 0.0 
       x_out_ptr(1:ccds,1:bigrows) = x_full_ptr(1:ccds,1:bigrows)
       y_out_ptr(1:ccds,1:ngoodcols,1:bigrows) = y_full_ptr(1:ccds,1:ngoodcols,1:bigrows)

       if (selecttimes) then
          call error ("NoSelectTimesYet","Selecttimes not supported yet!")
          write(errstr,*)"allocating ntimes(",ccds,")"
          call message(AppMsg,Sparse)
          allocate (ntimes(ccds))

          ntimes = count( ((x_out_ptr >= start) .AND. (x_out_ptr <= stop)), DIM=2 )
          goodrows = maxval(ntimes)

!          write(errstr,*)"allocating x_out_ptr(",ccds,",",goodrows,")"
!          call message(AppMsg,Sparse)
!          allocate (x_out_ptr(ccds,goodrows))
!          x_ptr = pack (timeptr,MASK = (timeptr >= start && timeptr <= stop),VECTOR = vect_x)

          do ccdno = 1, ccds
             write(errstr,*)"ntimes(",ccdno,")= ",ntimes(ccdno)
             call message(AppMsg,Sparse)
          end do
       end if


    else
       bigrows = goodrows
       write(errstr,*)"allocating x_out_ptr(",ccds,",",goodrows,")"," y_out_ptr(",ccds,",",ngoodcols,",",goodrows,")"
       call message(AppMsg,Sparse)
       allocate (x_out_ptr(ccds,goodrows),y_out_ptr(ccds,ngoodcols,goodrows))
!       x_out_ptr = HUGE(1.0)          ! put extra points off the graph
       x_out_ptr = 0.0  
       call message(AppMsg,Sparse,&
            "Plotting FRAME Number on x-axis (no OBT conversion)")
       do ccdno = 1 , ccds
          x_out_ptr(ccdno,1:n_ccd_row(ccdno)) = allx_ptr(ccdno,1:n_ccd_row(ccdno))
          y_out_ptr(ccdno,1:ngoodcols,1:n_ccd_row(ccdno)) = ally_ptr(ccdno,1:ngoodcols,1:n_ccd_row(ccdno))
          xmin_ptr(ccdno) = minval(x_out_ptr(ccdno,1:n_ccd_row(ccdno)))
          xmax_ptr(ccdno) = maxval(x_out_ptr(ccdno,1:n_ccd_row(ccdno)))
       end do

    end if

    ninrows = bigrows   ! reset ninrows 
    do i = 1 , ccds
       write(errstr,*)"i = ",i,"n_ccd_rows = ",n_ccd_row(i)," bigrows = ",bigrows,"xmin=",xmin_ptr(i),"xmax=",xmax_ptr(i)
       call message(AppMsg,Sparse)
    end do

  end subroutine get_arrays_frame
  !--------------------------------------------------------------------
  subroutine get_arrays_other(in_tab, ninrows, ngoodcols , &
                             xcolumnname, ycolrowptr, xcolunit, &
                             x_out_ptr, y_out_ptr , n_ccd_row , &
                             xmin_ptr , xmax_ptr)
    !------------------------------------------------------------------
    !
    ! subroutine to return a single precision array of
    ! values to use as the x-axis of the plot. Array can 
    ! be TIME or other. This subroutine calls OAL_ertToTimeTag
    ! if the selected x column is TIME, but in the input file
    ! the TIME column is type STRING.
    !
    ! in  : in_tab,ninrows,ngoodcols,xcolumnname,ycolrowptr
    ! out : xcolunit,x_out_ptr,y_out_ptr,n_ccd_row

    character(len=80), intent(out) :: xcolunit
    character(len=*), intent(in) :: xcolumnname
    character(len=19), allocatable, dimension(:) :: temp_string
    integer, intent(in) :: ninrows , ngoodcols
    integer :: ccdno
    integer(kind=Int32), dimension(:), pointer :: temp_int32 , n_ccd_row
    integer(kind=Int16), dimension(:), pointer :: temp_int16
    integer(kind=Int8), dimension(:), pointer :: temp_int8
!    logical(kind=bool),   dimension(:), pointer :: temp_bool
    real(kind=double), dimension(:), pointer :: temp_real64, &
                                                xmin_ptr , xmax_ptr
    real(kind=single), dimension(:), pointer :: temp_real32 , temp_x 
    real(kind=double), dimension(:,:), pointer :: x_out_ptr, ycolrowptr
    real(kind=double), dimension(:,:,:), pointer :: y_out_ptr
    integer :: xcoltype , i

    type(TableT)  :: in_tab
    type(ColumnT) :: in_col

    in_col = column(in_tab,trim(xcolumnname), READ)
    xcoltype = columnDataType(in_col)
    xcolunit = Units(in_col)
    
    allocate (x_out_ptr(ccds,1:ninrows))
    allocate (temp_x(1:ninrows))
    allocate (xmin_ptr(ccds), xmax_ptr(ccds))
    allocate (y_out_ptr(ccds,1:ngoodcols,1:ninrows))
    allocate (n_ccd_row(ccds))

    select case (xcoltype)
       !       case(Boolean)
       !          temp_bool => booleanData(in_col)
       !          x_out_ptr = real(temp_bool) 
    case(UInteger32)
       temp_int32 => uint32Data(in_col)
       temp_x(1:ninrows) = real(temp_int32(0:ninrows-1)) 
    case(UInteger16)
       temp_int16 => uint16Data(in_col)
       temp_x(1:ninrows) = real(temp_int16(0:ninrows-1)) 
    case(Integer32)
       temp_int32 => uint32Data(in_col)
       temp_x(1:ninrows) = real(temp_int32(0:ninrows-1)) 
    case(Integer16)
       temp_int16 => int16Data(in_col)
       temp_x(1:ninrows) = real(temp_int16(0:ninrows-1)) 
    case(Integer8)
       temp_int8 => int8Data(in_col)
       temp_x(1:ninrows) = real(temp_int8(0:ninrows-1)) 
    case(Real32)
       temp_real32 => real32Data(in_col)
       temp_x(1:ninrows) = real(temp_real32(0:ninrows-1))
    case(Real64)
       temp_real64 => real64Data(in_col)
       temp_x(1:ninrows) = real(temp_real64(0:ninrows-1))
    case(String)
       if (timeyes) then
          call message(AppMsg,Verbose,&
               "Converting String TIME column to Real TimeTag...")
          allocate (temp_string(ninrows))  ! could be YYYY-MM-DDTHH:MM:SS
          do i = 0,ninrows-1
             temp_string(i+1) = trim(stringCell(in_col,i))
             temp_x(i+1) =  OAL_ertToTimeTag(temp_string(i+1))
          end do
          deallocate(temp_string)
       else
          call error("XColStringNotTIME", &
               "String Type not supportable for non TIME column")
       end if
    case DEFAULT
       call error("TypeError",&
            "Column Type selected not supported for plotting.")
    end select

    call release(in_col)

    ! no way of converting some non-time oriented x-axis column to 
    ! GTIs, so even if mistakenly entered GTI filename, warn then do
    ! NOT plot GTIs at bottom of page...

    do ccdno = 1, ccds
       x_out_ptr(ccdno,1:ninrows) = temp_x(1:ninrows)
       y_out_ptr(ccdno,1:ngoodcols,1:ninrows) = ycolrowptr(1:ngoodcols,1:ninrows)
       xmin_ptr(ccdno) = minval(temp_x)
       xmax_ptr(ccdno) = maxval(temp_x)
    end do

    n_ccd_row = ninrows   ! for now...

    deallocate(temp_x)

    if (gtiyes) then
       if (.not. timeyes) then
          call warning("GTINotWithTimeX",&
               "GTIs inappropriate to non-time x-axis, GTIs will not be plotted.")
          gtiyes = .FALSE.
       end if
    end if

  end subroutine get_arrays_other

  !----------------------------------------------------------------------
  subroutine get_gti(gtifile, ccd_ptr, gti_ccd_ptr , n_gti_row , maxrows)
    !--------------------------------------------------------------------
    ! subroutine to open and read the GTI file block requested
    !
    ! gtifile  : Name of the input gti file
    ! gti_set   : Handle to the input gti file
    ! gti_tab   : Handle to the input gti table
    ! n_gti_row : number of rows in each ccd's GTI extension

    character(len=*)  :: gtifile
    character(len=2)  :: gtino
    character(len=80) :: datatype , gtiext
    integer(kind=Int16), dimension(:), pointer :: ccd_ptr
    integer(kind=Int32),  dimension(:), pointer :: n_gti_row
    real(kind=double),    dimension(:,:,:), pointer :: gti_ccd_ptr
    real(kind=double),    dimension(:), pointer :: temp_real64
    integer :: ccdno , maxrows

    type(DataSetT)  :: gti_set
    type(TableT)    :: gti_tab
    type(ColumnT)   :: in_col

    ! open gti file and read gti extension table

    write (errstr,*) "Opening GTI set: ",trim(gtifile)," READONLY"
    call message(AppMsg,Verbose)

    gti_set = dataSet(gtifile, Read)

    write (errstr,*) "Checking GTI extension(s): ",trim(gtiext)
    call message(AppMsg,Noisy)

    allocate (n_gti_row(ccds))

    ! first check the integrity of the gti file, set n_gti_row(ccdno)

    do ccdno = 1, ccds

       write (unit = gtino , FMT = '(I2.2)') ccd_ptr(ccdno)
       gtiext = trim(gtidatatype) // trim(gtino)

       if (hasBlock(gti_set, trim(gtiext))) then
          gti_tab = table(gti_set, gtiext)
          if ( hasAttribute(gti_tab,dtpar) ) then
             datatype = trim(stringAttribute(gti_tab,dtpar))
             if (index(datatype,trim(gtidatatype)) == 0) then
                call warning("DatatypeNotSTDGTI",&
                     "GTI extension DATATYPE not STDGTInn")
             end if
             n_gti_row(ccdno) = numberOfRows(gti_tab)
          else
             call error("NoDatatypeAttrib",&
                  "GTI file extension does not contain DATATYPE attribute")
          end if
          call release(gti_tab)
       end if

    end do

    ! do it again, and fill the gti_ccd_ptr array (saves space)

    maxrows = maxval(n_gti_row)
    allocate(gti_ccd_ptr(ccds,1:2,1:maxrows))
    gti_ccd_ptr = -1.

    do ccdno = 1, ccds

       write (unit = gtino , FMT = '(I2.2)') ccd_ptr(ccdno)
       gtiext = trim(gtidatatype) // trim(gtino)
       write (errstr,*) "Reading GTI extension: ",trim(gtiext)
       call message(AppMsg,Verbose)
       
       gti_tab = table(gti_set, gtiext) ! get extension STDGTInn
       
       if (hasColumn(gti_tab, trim(startcolpar)) .AND. &
            hasColumn(gti_tab, trim(stopcolpar))) then
             
          ! get the START column (will be DOUBLE precision)

          in_col  = column(gti_tab,startcolpar, READ)
          temp_real64 => real64Data(in_col)
          gti_ccd_ptr(ccdno,1,1:n_gti_row(ccdno)) = real(temp_real64)
          call release(in_col)
             
          ! get the STOP column

          in_col  = column(gti_tab,stopcolpar, READ)
          temp_real64 => real64Data(in_col)
          gti_ccd_ptr(ccdno,2,1:n_gti_row(ccdno)) = real(temp_real64)
          call release(in_col)   
          !             if ( HasAttribute(gti_tab,"TSTART") ) then
          !                tstartgti = real32Attribute(gti_tab,"TSTART")
          !             else
          !                write(errstr,*) "Attribute TSTART missing in input GTI file. Using START column row 1"
          !                call warning("missingATTRIB")
          !                tstartgti = real(gtistartptr(0))
          !             end if
          !             
          !             if ( HasAttribute(gti_tab,"TSTOP") ) then
          !                tstopgti = real32Attribute(gti_tab,"TSTOP")
          !             else
          !                write(errstr,*) "Attribute TSTOP missing in input GTI file. Using STOP column row ",ngtirows
          !                call warning("missingATTRIB")
          !                tstopgti = real(gtistopptr(ngtirows))
          !             end if
       else
          write(errstr,*) "GTI extension: ",ccdno, " does not contain START/STOP Columns"
          call warning("NoSTARTSTOPCols")
       end if
       call release(gti_tab)       
    end do

    call release(gti_set)

  end subroutine get_gti

  !-----------------------------------------------------------------------
  function change_case(instring,dir) result(outstring)
    !---------------------------------------------------------------------

    ! Switch from lower to upper case or vice-versa: From Jean Ballet
    ! instring : input string
    ! dir      : direction (1: to lower case; 2: to uppercase; 0: switch case)
    
    character(len=*), intent(in) :: instring
    integer, intent(in) :: dir
    character(len=len(instring)) :: outstring
    
    ! ASCII positions of lower case and upper case letters
    
    integer, parameter :: start_up  = iachar("A"), end_up  = iachar("Z"), &
         start_low = iachar("a"), end_low = iachar("z"), &
         offs = start_low - start_up
    
    integer :: length, icur, pos
    
    length = len(instring)
    outstring = instring
    do icur = 1, length
       pos = iachar(instring(icur:icur))
       if (pos >= start_up .and. pos <= end_up .and. dir < 2) then
          pos = pos + offs
       else if (pos >= start_low .and. pos <= end_low .and. dir /= 1) then
          pos = pos - offs
       end if
       outstring(icur:icur) = achar(pos)
    enddo
    
  end function change_case

  !----------------------------------------------------------------------
  subroutine create_main_labels(infile, gtifile)
    !--------------------------------------------------------------------
    ! Generates labels to add on plot.
    !
    character(len=*), intent(inout) :: infile, gtifile
    character(len=128) :: cmjdref , cduration
    !
    ! check for keywords (attributes) in infile to put as plot title 
    ! don't add keyword to string if not present so at least a partial
    ! title set can be constructed
    ! Title to plot should look like this:
    !                       HKAUXPLOT output
    !          HK file = 0000_0004930000_M1S00071HK.FIT
    !        GTI file = 0000_0004930000_M1S00071GTI.FIT
    !               XMM - EMOS1 - OBS_ID = 123456
    !   START: 2000-02-19T13:36:21 - STOP: 2000-02-19T13:36:21
    !         DURATION: 55283.0 (s) - MJDREF = 51544.0 
    !                     OBJECT: Crab Nebula
    !

    labell(1) = 'HKAUXPLOT Output'
    labell(2) = 'Input File = '// trim(infile)

    if (.not. gtiyes) then
       gtifile = "None Selected"
    end if

    labell(3) = 'GTI File = '// trim(gtifile)
    labell(4) = trim(telescop) // ' - ' // trim(instrume) // ' - ' // &
         trim(obs_id) // ' - ' // trim(exp_id)
    labell(5) = 'START : ' // trim(date_obs) // ' - ' // &
               'END : ' // trim(date_end)

    write(cduration,'(F8.1)') duration
    write(cmjdref,'(F8.1)') mjdref
    labell(6) = 'DURATION: ' // trim(cduration) // ' - ' // &
             'MJDREF: ' // trim(cmjdref)

    labell(7) = 'OBJECT: ' // trim(object)

  end subroutine create_main_labels


  !----------------------------------------------------------------------
  subroutine do_plot(device, outplot, xcolname, xcolunit, ycolsptr , &
                     yunitsptr, ptpp, offs, nycols, nrows, &
                     n_ccd_row , y_out_ptr, x_out_ptr, ccd_ptr, &
                     gti_array)
    !--------------------------------------------------------------------
    ! Create the actual PGPLOT ps file of the data
    !  
    ! declare "inbound" stuff first
  
    integer, intent(in) :: nycols , nrows
    integer, intent(inout) :: ptpp
    character(len=*), intent(inout) :: device , outplot
    character(len=*), dimension(:), pointer :: ycolsptr, yunitsptr
    character(len=*), intent(in) :: xcolname , xcolunit
    integer(kind=Int32),  dimension(:), pointer :: n_ccd_row
    integer(kind=Int16),  dimension(:), pointer :: ccd_ptr
    real(kind=double), dimension(:,:), pointer :: x_out_ptr, gti_array
    real(kind=double), dimension(:,:,:), pointer :: y_out_ptr
    real(single), allocatable, dimension(:) :: temp_x , temp_y , &
                                               temp_gti , txl
    real(single), allocatable, dimension(:,:) :: intv_x , intv_gti
    real(single), allocatable, dimension(:,:,:) :: intv_y
    character(len=3) :: xwinstring = 'XW'
    character(len=5) :: ccdchar
    integer :: i, k, l, remrows, pplotted , panl, curplot , &
         low , cno , totplots , pages , pn_pp , page , col
    real ::  xlo(ccds), xhi(ccds), xl, xh, ylo , yhi , factor, &
             currval , preval , realpages , xmin , xmax , ymin, ymax, &
             realintvals , realtotpn , glo , ghi
    character(len=128) :: outdevice
    integer :: PGBEG , maxpnpp , calcpnpp , loop , curpanel , modo, &
               extra , intervals , rempts , totpn , rempn  , curintval, &
               panonpg , pnPERint, pnINintval , first , last
    logical :: exists , offs , new , didit

    exists = .FALSE. 

    ! calculate the page layout

    ! e.g if ptpp (selected) = 600 (default) and nrows is 300, then
    ! ptpp (points per page) becomes 300

    ptpp = min (ptpp,nrows)         ! minimum of selected pts per pg & nrows
    realintvals = real(nrows)/real(ptpp)  ! real number of x-axis intervals

    if ((realintvals - real(INT(realintvals))) > 0.) then
       intervals = int(realintvals) + 1
    elseif (NINT(realintvals) == 0) then
       intervals = int(realintvals) + 1
    else
       intervals = int(realintvals)
    end if

    realtotpn = real(nycols) * real(ccds) * real(intervals) ! panels needed

    ! total panels

    if ((realtotpn - real(INT(realtotpn))) > 0.) then
       totpn = int(realtotpn) + 1
    elseif (NINT(realtotpn) == 0) then
       totpn = int(realtotpn) + 1
    else
       totpn = int(realtotpn)
    end if

    rempts = mod(nrows,ptpp)       ! 'excess' points above integer number
    maxpnpp = 20                   ! maximum panels per page + title + bottom
    pnPERint = (nycols * ccds)     ! actual panels per INTERVAL
    if (pnPERint > 18) then
       call warning("PlotSpanPages",&
            "More than 18 panels per page; intervals will span more than one page")
    end if
    calcpnpp = pnPERint + 2        ! 2 extra panels for title and bottom
    pn_pp = min(maxpnpp,calcpnpp)  ! actual panels per PAGE
    realpages = real(totpn) / real(pnPERint)
    rempn = mod(totpn,pn_pp)       ! remaining panels at end of series

    if (ABS(realpages - real(INT(realpages))) > 0.) then
       pages = int(realpages) + 1
    elseif (NINT(realpages) == 0) then
       pages = int(realpages) + 1
    else
       pages = int(realpages)
    end if

    write(errstr,*) "total points: ",nrows," x-points per page: ",ptpp
    call message(AppMsg,Verbose)
    if (nrows < 3 .OR. ptpp < 3 ) then
       call error ("TooFewPoints",&
            "Less than 3 points per page -- no plot generated")
    end if

    write(errstr,*) "x-intervals=",INT(realintvals),&
         "plus 1 with:",rempts,"points. Total intervals= ",intervals
    call message(AppMsg,Verbose)
    write(errstr,*)"pages = ",pages," with ",pn_pp," TOTAL panels per page"
    call message(AppMsg,Verbose)
    write(errstr,*) "panels per interval= ",pnPERint,&
         " panels remaining before new page: ",rempn
    call message(AppMsg,Verbose)
    write(errstr,*) " total panels: ",totpn
    call message(AppMsg,Verbose)

    if (selectccds) then
       write(errstr,*)"CCDs selected=",ccds,"ccds=",ccd_ptr(1:ccds)
       call message(AppMsg,Verbose)
    end if

    write(errstr,*)"ncols selected=",nycols,"cols= ",ycolsptr(1:nycols)(1:10)
    call message(AppMsg,Verbose)
    do i = 1 , ccds
       write(errstr,*)"n_ccd_row(",i,") = ",n_ccd_row(i)
       call message(AppMsg,Verbose)
    end do

    if (device(1:1) /= '/') device= "/" // trim(device)

    ! if device is PS, change it to /PS -- fix typo on command line
    ! also, put plot in 'vertical' (portrait) mode 
    ! if device is xw go with it in landscape mode

    if (index(device,trim(xwinstring)) == 0) then
       if (selectoutfile) then
          outdevice = trim(outplot) // trim(device)
          inquire(file=trim(outplot),exist=exists)
          if (exists) then
             if ( clobber() ) then
                write(errstr,*) "File ",trim(outplot),&
                     " exists. Will be overwritten"
                call warning("OutfileExistsOverwriting")
             else
                write(errstr,*) "File ",trim(outplot), &
                     " exists. Will NOT be overwritten, SAS_CLOBBER not set"
                call error("OutfileExistsNOoverwrite")
             end if
          end if
       else
          outdevice=device
       end if
    else  
       outdevice=device
    end if

    !
    ! start pgplot with pgbeg call with, e.g. outdevice="plot.ps/ps" 
    !

    if ( pgbeg(0,trim(outdevice),1,1) /= 1 ) then
       write(errstr,*) "invalid device, ",trim(outdevice)," selected"
       call error ("invalidDevice")
    end if

    ! determine the character height adjustment factor depending
    ! on how many columns are requested

    factor = 0.5
    !
    ! these needs work! 
    !
    if (pn_pp <= 2) then
       factor = factor + 0.5
    elseif (pn_pp <=4) then
       factor = factor + 1.3
    elseif (pn_pp <=6) then
       factor = factor + 1.7
    elseif (pn_pp <=8) then
       factor = factor + 2.5
    elseif (pn_pp <=10) then
       factor = factor + 3.0
    elseif (pn_pp <=12) then
       factor = factor + 3.3
    elseif (pn_pp <=14) then
       factor = factor + 3.6
    elseif (pn_pp <=16) then
       factor = factor + 4.0
    elseif (pn_pp <=18) then
       factor = factor + 4.2
    elseif (pn_pp >18 ) then
       factor = factor + 4.7
    end if
    
    remrows = nrows
    panl = 1
    pplotted = 0
    xl = 0.
    xh = 1.
    ylo = 0.
    yhi = 1.
    col = 1
    cno = 1
    loop = 1
    xmin = 1e7
    xmax = 1e8
    ymin = 0.
    ymax = 50.
    page = 1
    curpanel = 1
    pnINintval = 1
    panonpg = 1

!    INTERVAL_Loop: do

    do curintval = 1 , intervals

       if (curpanel > totpn) exit

       write(errstr,*)"page=",page," of ",pages," curintval=",&
            curintval," of ",intervals
!       call message(AppMsg,Verbose)

       first = ((curintval - 1) * ptpp) + 1
       last = curintval*ptpp

       if (curintval == intervals) then 
          if (rempts > 0) then
             ptpp = rempts   ! use 'excess' pts if in last interval
             last = first + ptpp - 1 
          end if
       end if

       write(errstr,*)"allocating intv_x(",ccds,",",ptpp,")"
       call message(AppMsg,Verbose)
       write(errstr,*)"allocating intv_y(",ccds,",",nycols,",",ptpp,")"
       call message(AppMsg,Verbose)
       write(errstr,*)"allocating temp_x(",ptpp,")"
       call message(AppMsg,Verbose)
       write(errstr,*)"allocating temp_y(",ptpp,")"
       call message(AppMsg,Verbose)
       ! 
       ! first, get an 'interval' worth of data.
       ! 
       allocate(intv_x(ccds,ptpp))          ! page's worth of x-axis values
       allocate(intv_y(ccds,nycols,ptpp))   ! page's worth of y-axis values
       allocate (temp_y(ptpp),temp_x(ptpp)) ! arrays to send to pgplot

       intv_x(1:ccds,1:ptpp) = real(x_out_ptr(1:ccds,first:last))
       intv_y(1:ccds,1:nycols,1:ptpp) = real(y_out_ptr(1:ccds,1:nycols,first:last))
       if (gtiyes) then
          write(errstr,*)"allocating intv_gti(",ccds,",",ptpp,")"
          call message(AppMsg,Verbose)
          write(errstr,*)"allocating temp_gti(",ptpp,")"
          call message(AppMsg,Verbose)
          allocate(intv_gti(ccds,ptpp),temp_gti(ptpp))
          intv_gti(1:ccds,1:ptpp) = gti_array(1:ccds,1:ptpp)
          glo = 0.
          ghi = 1.
       end if

       ! get page min and max x-values for plot bounds                
                
       xlo = minval(intv_x, 2)
       xhi = maxval(intv_x, 2)

       xl = minval(xlo)
       xh = maxval(xhi)

       xl = (xl + xl)  - xl    ! kludge to get rid of very small values
       xh = (xh + xh)  - xh    ! (e.g. e-18) and set to 0.

       write(errstr,'(A,I8,A,I8)')&
            "xlo=",INT(xl),"xhi=",INT(xh)
!       call message(AppMsg,Verbose)
       !
       ! fix x
       !
       if (curintval == intervals) then    ! last set thru
          if (rempts > 0) then
             allocate (txl(ccds))
             do i = 1 , ccds
                txl(i) = intv_x(i,1)
             end do
             xl = minval(txl)
             do i = 1 , ccds
                txl(i) = intv_x(i,ptpp)
             end do
             xh = maxval(txl)
             write(errstr,*) "xl,xh= ",xl,xh
!             call message(AppMsg,Verbose)
             deallocate (txl)
          end if
       end if

       Panel_Loop: do

          do cno = 1 , ccds

             temp_x(1:ptpp) = intv_x(cno,1:ptpp)

             if (gtiyes) then
                write(errstr,*) "filling temp_gti(1:",ptpp,"), cno = ",cno
                call message(AppMsg,Sparse)
                temp_gti(1:ptpp) = intv_gti(cno,1:ptpp)
             end if

             do col = 1 , nycols

                temp_y(1:ptpp) = real(intv_y(cno,col,1:ptpp))

                ylo = minval(temp_y)
                yhi = maxval(temp_y)

!                temp_y = (temp_y + yhi)  - yhi ! kludge to round numbers off
                
                ylo = minval(temp_y)
                yhi = maxval(temp_y)

                if (ylo == yhi) then
                   if (ylo == 0.) yhi = 1.      ! kludge city for all 0's!!!
                   ylo = 0.
                end if

                write(errstr,'(A5,I8,A5,I8,A5,I7,A5,I7)')&
                     "xlo=",INT(xl),"xhi=",INT(xh),&
                     "ylo=",INT(ylo),"yhi=",INT(yhi)
!                call message(AppMsg,Verbose)

                ! for each new page...

                if (panonpg == 1) then 
                   call pgiden             ! put date and ident at bottom
                   call pgsubp(1,pn_pp)    ! columns + title
                   call pgpanl(1,1)        ! go to panel 1 for title
                   call add_top(factor)       ! add title
                   call pgpage                     ! start new panel
                   panonpg = 2
                   new = .true.
                end if
                !
                ! write actual data here
                !

                if (gtiyes) then
                   call pgsvp(0.1,0.90,0.3,1.0)
                else 
                   call pgsvp(0.1,0.90,0.05,1.0)
                end if

                call pgswin(xl,xh,ylo,yhi)

                if ((xl > 10000.) .OR. (xh > 10000.)) then
                   call pgbox('BCT2',0.0,0,'BCNTV',0.0,3)
                else 
                   call pgbox('BCT',0.0,0,'BCNTV',0.0,3)
                end if

                if (aux) then
                   call pgbin (ptpp,temp_x,temp_y,.true.)
!                   call pghist (ptpp,temp_x,ylo,yhi,4,2)
!                   call pgpoly (ptpp,temp_x,temp_y)
!                   call pgpt (ptpp,temp_x,temp_y,16)
                else
                   call pgline (ptpp,temp_x,temp_y)
                end if
                !
                ! write unit string to left, and columnname to
                ! write, plus any ccdid on right also.
                !
                call pgmtxt('R' , 2. , 0.5 , 0.5 , trim(ycolsptr(col)))
                if (aux) then 
                   write (unit=ccdchar,fmt='(A3,I2.2)')"CCD_",ccd_ptr(cno)
                   call pgmtxt('RV' , 1.5 , 0.05 , 0.5 ,trim(ccdchar))
                end if

                call pgmtxt('L' , 3.5 , 0.5 , 0.5 , trim(yunitsptr(col)))

                if (gtiyes) then

                   write(errstr,*) "got to here"
                   call pgswin(xl,xh,0.,1.)
                   call pgsvp(0.1,0.90,0.05,0.3)
                
                   ! If it is the last panel on the page, write the x-axis
                   ! values at the bottom of the page (plus tick marks).
                   ! If at last CCD and Column of an interval, plot
                   ! the x-axis units at last column. Will then panel down
                   ! to end of current page and start new one.
                
                   call pgbox('BCT',0.0,0,'BCTV',1.0,1)
                   call pgline (ptpp,temp_x,temp_gti)
          
                   ! once the GTI interval is plotted, fill in each area of the
                   ! panel where GTI value is 1 (eg. when the plot is IN a GTI.
                
                   ylo = minval(temp_gti) 
                   yhi = maxval(temp_gti)
                   call pgsfs(3)           ! hatched fill area for GTI panel
                   call pgshs(45.,10.,0.0) ! 45 degree angle, 1/10th separation
                
                   if (ylo == yhi) then
                      if (ylo > 0.) then  ! if panel contains all "on" values:
                                          ! fill entire panel
                         call pgrect(temp_x(1),temp_x(ptpp),0.,1.)
                      end if
                   else
                      !
                      ! loop thru and fill all "on" intervals in a panel
                      ! get first value of interval (0 or 1)
                      currval = temp_gti(1)
                      preval = temp_gti(1)
                      low = 1
                      k = 1
             
                      ! this loop creates and fills the boundaries of a GTI
                      ! interval. The interval is 'filled' with 45 degree
                      ! hashed lines.
                
                      GTI_Loop: do
                         if (k == ptpp) then
                            if (temp_gti(ptpp) > 0.) then
                               call pgrect(temp_x(low),temp_x(ptpp),ylo,yhi)
                            end if
                            exit GTI_Loop
                         end if
                         if (temp_gti(k) /= currval) then
                            if (preval > 0.) then
                               call pgrect(temp_x(low),temp_x(k-1),ylo,yhi)
                               currval = temp_gti(k)
                            end if
                            low = k
                         end if
                         preval = temp_gti(k)
                         k = k + 1
                      end do GTI_Loop
                   end if
             
                   call pgmtxt('LV' , 1.6 , 0.5 , 0.5 ,"on/off")
                   write (unit=ccdchar,fmt='(A3,I2.2)')"GTI",ccd_ptr(cno)
                   call pgmtxt('RV' , 1.5 , 0.5 , 0.5 ,trim(ccdchar))
                end if
                !
                ! if still some panels left in interval, but get to
                ! bottom of page, write text and start new page, then
                ! finish remaining panels within interval
                !
                didit = .false.
                write (errstr,*)"intpn=",pnINintval,"pnpg=",&
                     panonpg,"pn_pp=",pn_pp,"pnp=",pnPERint,"cno=",cno,"col=",col
!                call message(AppMsg,Verbose)

                if (panonpg == (pn_pp-1)) then
                   write (errstr,*) panonpg,"pn_pp=",pn_pp-1,&
                        "pnp=",pnPERint,"cno=",cno,"col=",col
!                   call message(AppMsg,Verbose)
                   if ((xl > 10000.) .OR. (xh > 10000.)) then
                      call pgbox('N2',0.0,0,' ',1.0,1)
                   else 
                      call pgbox('N',0.0,0,' ',1.0,1)
                   end if
!                   call pgbox('N',0.0,0,' ',1.0,1)
                   call add_bottom(offs,xcolname,xcolunit,page,pages)
                   call pgpage      ! now at end of page
                   call pgpage      ! start new page
                   page = page + 1
                   panonpg = 0
                   didit = .true.
                end if
                !
                ! if panel in interval = panels per interval expected,
                ! add bottom text, panel down to bottom of page, then
                ! start new page.
                !
                if (pnINintval == pnPERint) then
                   if (rempn > 0) then 
                      if (.not. didit) then 
                         if ((xl > 10000.) .OR. (xh > 10000.)) then
                            call pgbox('N2',0.0,0,' ',1.0,1)
                         else 
                            call pgbox('N',0.0,0,' ',1.0,1)
                         end if
                         call add_bottom(offs,xcolname,xcolunit,page,pages)
                         do i = panonpg  , (pn_pp-1)
                            call pgpage
                         end do
                         call pgpage      ! now at end of page
                         call pgpage      ! now at start of new page
                         page = page + 1
                      end if
                   end if
                   panonpg = 1
                   pnINintval = 1
!                   curpanel = curpanel + 1        ! 'real' total panels done
                   exit Panel_Loop
                end if
                call pgpage                    ! increment panel on page
                panonpg = panonpg + 1          ! panel on page
                curpanel = curpanel + 1        ! 'real' total panels done
                pnINintval = pnINintval + 1    ! panel within interval
             end do
          end do
       end do Panel_Loop
       deallocate (intv_x,intv_y,temp_x,temp_y)
       if (gtiyes) then 
          deallocate (intv_gti,temp_gti)
       end if
    end do

    ! end the plot
    call message(AppMsg,Verbose,"Closing plot with pgend")
    call pgend
    
  end subroutine do_plot

  !----------------------------------------------------------------------
  subroutine add_top(factor)
    !--------------------------------------------------------------------

    real(single), intent(in) :: factor
    integer :: i

    call pgpanl(1,1)
    call pgsch (factor)

    do i=1,7
       call pgmtxt('T' , -(1.0 + (i-1)*0.9) , 0.5 , 0.5 , trim(labell(i)))
    end do

    call pgsch (factor)

  end subroutine add_top

  !----------------------------------------------------------------------
  subroutine add_bottom(offs,xcolname,xcolunit,page,pages)
    !--------------------------------------------------------------------

    logical :: offs
    character(len=*), intent(in) :: xcolname , xcolunit
    character(len=256) :: str
    integer, intent(in) :: page, pages

    if (timeyes) then
       if (offs) then
          write(str,'(A,F12.3)')'Time (s) since TSTART =  ',tstartin
       else 
          write(str,'(A,F12.3)')'Time (s) since MJDREF =  ',mjdref
       end if
    else
       write(str,*)trim(xcolname),' ',trim(xcolunit)
    end if

    call pgmtxt('B' , 3.5 , 0.5 , 0.5 ,trim(str))

    ! write page number at bottom

    write(str,'(A,I2.2,A,I2.2)')'Page ',page,"/",pages
    call pgmtxt('B' , 5.5 , 0.5 , 0.5 ,trim(str))

  end subroutine add_bottom

  !----------------------------------------------------------------------
  subroutine check_ss(start,stop,type)
    !--------------------------------------------------------------------

    real (single) start, stop
    integer type
    
    SELECT CASE (type)
       
    CASE (0)   ! selecting on ROWS
       call message(AppMsg,Verbose,"Selecting infile on ROWS.")
       if (int(start) >= int(stop)) then
          write (errstr,*) "Start row selected, ",int(start),&
               " .GE. stop row selected, ",int(stop)
          call error("RowRangeInvalid")
       end if
       if (int(start) <= 0) then
          write (errstr,*) "Start row selected, ",int(start),&
               ".LE. 0"
          call error("RowRangeInvalid")
       end if
       if (int(stop) <= 0) then
          write (errstr,*) "Stop row selected, ",int(stop),&
               ".LE. 0"
          call error("RowRangeInvalid")
       end if
    CASE (1) ! selecting on TIME
       call message(AppMsg,Verbose,"Selecting infile on TIME.")
       if (start >= stop) then
          write (errstr,*) "Start time selected, ",start,&
               " .GE. stop time selected, ",stop
          call error("TimeRangeInvalid")
       end if
       if (start <= 0.) then
          write (errstr,*) "Start time selected, ",start, &
               ".LE. 0"
          call warning("StartTimeLTZero")
       end if
       if (stop <= 0.) then
          write (errstr,*) "Stop time selected, ",stop, &
               ".LE. 0"
          call warning("StopTimeLTZero")
       end if
    CASE (2) ! selecting on FRAMES
       call message(AppMsg,Verbose,"Selecting infile on FRAMES.")
       if (int(start) >= int(stop)) then
          write (errstr,*) "Start frame selected, ",int(start),&
               " .GE. stop frame selected, ",int(stop)
          call error("FrameRangeInvalid")
       end if
       if (int(start) <= 0) then
          write (errstr,*) "Start frame selected, ",int(start),&
               ".LE. 0"
          call error("FrameRangeInvalid")
       end if
       if (int(stop) <= 0) then
          write (errstr,*) "Stop frame selected, ",int(stop),&
               ".LE. 0"
          call error("FrameRangeInvalid")
       end if
    CASE DEFAULT
       call error ("IllegalType","Illegal Type Code in check_ss")      
    END SELECT

  end subroutine check_ss

end module hkauxplot_mod
